import logging
from pathlib import Path

import click

from zamba.models.manager import ModelManager

default_model_dir = Path(__file__).parent / 'models' / 'assets'
default_model_dir.mkdir(exist_ok=True)


@click.group()
def main():
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')


@main.command()
@click.argument('data_path',
                type=click.Path(exists=True,
                                file_okay=True,
                                dir_okay=True),
                default=Path('.'))
@click.argument('pred_path',
                type=click.Path(),
                default=Path('.', 'output.csv'))
@click.option('--tempdir',
              type=click.Path(exists=True),
              default=None,
              help="Path to temporary directory. If not specified, OS temporary directory is used.")
@click.option('--proba_threshold',
              type=float,
              default=None,
              help="Probability threshold for classification. if specified binary predictions are returned with 1 "
                   "being greater than the threshold, 0 being less than or equal to. If not specified, probabilities "
                   "between 0 and 1 are returned.")
@click.option('--output_class_names',
              is_flag=True,
              default=False,
              help="If True, we just return a video and the name of the most likely class. If False, "
                   "we return a probability or indicator (depending on --proba_threshold) for every "
                   "possible class.")
@click.option('--model_profile',
              type=str,
              default='full',
              help="Defaults to 'full' which is slow and accurate; can be 'fast' which is faster and less accurate.")
@click.option('--weight_download_region',
              type=str,
              default="us",
              help="Defaults to 'us', can also be 'eu' or 'asia'. Region for server to download weights.")
@click.option('--verbose',
              is_flag=True,
              default=False,
              help="Displays additional logging information during processing.")
def predict(data_path,
            pred_path,
            tempdir,
            proba_threshold,
            output_class_names,
            model_profile,
            weight_download_region,
            verbose):
    """Identify species in a video.

      This is a command line interface for prediction on camera trap footage. Given a path to camera trap footage,
      the predict function use a deep learning model to predict the presence or absense of a variety of species of
      common interest to wildlife researchers working with camera trap data.

    """

    click.echo(f"Using data_path:\t{data_path}")
    click.echo(f"Using pred_path:\t{pred_path}")

    # Load the model into manager
    manager = ModelManager(model_path=default_model_dir,
                           model_class='cnnensemble',
                           proba_threshold=proba_threshold,
                           tempdir=tempdir,
                           output_class_names=output_class_names,
                           model_kwargs=dict(profile=model_profile,
                                             download_region=weight_download_region))

    # Make predictions, return a DataFrame
    manager.predict(data_path, pred_path=pred_path, save=True)


@main.command()
@click.argument('data_path',
                type=click.Path(exists=True,
                                file_okay=True,
                                dir_okay=True))
@click.argument('labels',
                type=click.File())
@click.option('--tempdir',
              type=click.Path(exists=True),
              help="Path to temporary directory. If not specified, OS temporary directory is used")
@click.option('--batch_size',
              default=8,
              type=click.IntRange(min=1,
                                  max=256,
                                  clamp=True),
              help="Batch size for model inputs, defaults to 8 samples.")
@click.option('--weights_out',
              type=click.Path(exists=True),
              help="Output path for saved weights.")
def tune(data_path, labels, tempdir, batch_size, weights_out):
    """ [NOT IMPLEMENTED] Update network with new data.

        Finetune the network for a different task by keeping the
        trained weights, replacing the top layer with one that outputs
        the new classes, and re-training for a few epochs to have the
        model output the new classes instead.

    """
    click.echo("Finetuning an algorithm to new data is not yet implemented.")


@main.command()
@click.argument('data_path',
                type=click.Path(exists=True,
                                file_okay=True,
                                dir_okay=True))
@click.argument('labels',
                type=click.File())
@click.option('--tempdir',
              type=click.Path(exists=True),
              help="Path to temporary directory. If not specified, OS temporary directory is used")
def train(data_path, labels, tempdir):
    """ [NOT IMPLEMENTED] Retrain network from scratch.

        Train the weights from scratch using
        the provided data_path and labels.
    """
    click.echo("Training an algorithm from scratch on new data is not yet implemented.")



from datetime import datetime
from enum import Enum, EnumMeta
import logging
from pathlib import Path

from zamba.models.model import SampleModel
from zamba.models.cnnensemble_model import CnnEnsemble


class GetItemMeta(EnumMeta):
    """Complicated override so that we can use hashing in ModelManager init.

    """
    def __getitem__(cls, name):
        for e in cls:
            if e.string == name:
                return e

        raise ValueError(f"Key '{name}' not in ModelName Enum.")


class ModelName(Enum, metaclass=GetItemMeta):
    """Allows easy control over which Model subclass to load. To add a new model class, add a line like ``NEW_MODEL
    = ('new_model', NewModelClass)``

        Args:
            string (str) : string used to reference the model model
            model (Model) : model class to instantiate into manager

    """

    WINNING = ('cnnensemble', CnnEnsemble)
    SAMPLE = ('sample', SampleModel)

    def __init__(self, string, model):
        self.string = string
        self.model = model

    def __eq__(self, other):
        return other == self.value


class ModelManager(object):
    """Mediates loading, configuration, and logic of model calls.

        Args:
            model_path (str | Path) : path to model weights and architecture
                Required argument. Will be instantiated as Model object.
            proba_threshold (float) : probability threshold for classification
                Defaults to ``None``, in which case class probabilities are returned.
            tempdir (str | Path) : path to temporary directory
                If specific temporary directory is to be used, its path is passed here. Defaults to ``None``.
            verbose (bool) : controls verbosity of prediction, training, and tuning methods
                Defaults to ``True`` in which case training, tuning or prediction progress will be logged.
            model_class (str) : controls whether sample model class or production model class is used
                Defaults to "winning". Must be "winning" or "sample".
    """
    def __init__(self,
                 model_path=Path('.'),
                 proba_threshold=None,
                 output_class_names=False,
                 tempdir=None,
                 verbose=False,
                 model_class='cnnensemble',
                 model_kwargs=dict()):

        self.logger = logging.getLogger(f"{__file__}")

        self.model_path = Path(model_path)
        self.model_class = ModelName[model_class].model

        self.tempdir = tempdir
        self.model = self.model_class(model_path, verbose=verbose, **model_kwargs)
        self.proba_threshold = proba_threshold

        self.output_class_names = output_class_names

        self.verbose = verbose

    def predict(self, data_path, save=False, pred_path=None):
        """
        Args:
            data_path (str | Path) : path to input data
            pred_path (str | Path) : where predictions will be saved

        Returns: DataFrame of predictions

        """

        data_path = Path(data_path)

        # cnn ensemble doesn't use simple data loader, samples do...
        if self.model_class != 'cnnensemble':
            data = self.model.load_data(data_path)
            preds = self.model.predict(data)
        else:
            preds = self.model.predict(data_path)

        # threshold if provided
        if self.proba_threshold is not None:
            preds = preds >= self.proba_threshold

        if self.output_class_names:
            preds = preds.idxmax(axis=1)

        if save:
            if pred_path is None:
                timestamp = datetime.now().isoformat()
                pred_path = Path('.', f'predictions-{data_path.parts[-1]}-{timestamp}.csv')

            preds.to_csv(pred_path)

            if self.verbose:
                self.logger.info(f"Wrote predictions to {pred_path}")

        if self.verbose or self.output_class_names:
            self.logger.info(preds)

        return preds

    def train(self):
        """

        Returns:

        """
        pass

    def tune(self):
        """

        Returns:

        """
        pass



import pickle
from pathlib import Path
from shutil import rmtree
import tempfile

import pandas as pd


try:
    from tensorflow.python import keras
except ImportError:
    msg = "Zamba must have tensorflow installed, run either `pip install zamba[cpu]` "\
          "or `pip install zamba[gpu]` " \
          "depending on what is available on your system."
    raise ImportError(msg)


class Model(object):
    """Abstract class implementing required methods for any model in the api.

        Args:
            model_path (str | Path) : path to model files
                Converted to ``Path`` object in ``__init__``. Defaults to ``None``.
            tempdir (str | Path) : path to temporary diretory
                Path to temporary directory, if used. Defaults to ``None``.

        Attributes:
            model_path (str | Path) : path to model files
                Converted to ``Path`` object in ``__init__``. Defaults to ``None``.
            tempdir (str | Path) : path to temporary diretory
                Path to temporary directory, if used. Defaults to ``None``.
            delete_tempdir (bool) : whether to clean up tempdir
                Clean up tempdir if used.

    """

    def __init__(self, model_path=None, tempdir=None, verbose=False):
        self.model_path = Path(model_path) if model_path is not None else None
        self.delete_tempdir = tempdir is None
        self.tempdir = Path(tempfile.mkdtemp()) if self.delete_tempdir else Path(tempdir)
        self.verbose = verbose

    def __del__(self):
        """ If we use the default temporary directory, clean this
            up when the model is removed.
        """
        if self.delete_tempdir:
            rmtree(self.tempdir)

    def predict(self, X):
        """

        Args:
            X: Input to model.

        Returns: DataFrame of class probabilities.

        """
        pass

    def fit(self, X, y):
        """Use the same architecture, but train the weights from scratch using
        the provided X and y.

        Args:
            X: training inputs
                Numpy arrays probably
            y: training labels
                Class labels

        Returns:

        """
        pass

    def finetune(self, X, y):
        """Finetune the network for a different task by keeping the
        trained weights, replacing the top layer with one that outputs
        the new classes, and re-training for a few epochs to have the
        model output the new classes instead.

        Args:
            X:
            y:

        Returns:

        """
        pass

    def save_model(self):
        """Save the model weights, checkpoints, to model_path.

        Returns:

        """


class SampleModel(Model):
    """Sample model for testing.

        Args:
            model_path:
            tempdir:
    """
    def __init__(self, model_path=None, tempdir=None, verbose=False):
        super().__init__(model_path, tempdir=tempdir)

        self.model = self._build_graph() if self.model_path is None else keras.models.load_model(self.model_path)
        self.verbose = verbose

    def _build_graph(self):
        """Simple keras graph for testing api.

        Takes two numbers, adds them, also multiplies them, outputs both results.

        Returns: keras model for testing

        """

        # build simple architecture to multiply two numbers
        w1 = keras.layers.Input(shape=(1,), name="w1")
        w2 = keras.layers.Input(shape=(1,), name="w2")

        add = keras.layers.add([w1, w2])
        mult = keras.layers.multiply([w1, w2])
        out = keras.layers.concatenate([add, mult])

        return keras.models.Model(inputs=[w1, w2], outputs=out)

    def predict(self, X):
        """

        Args:
            X (list | numpy array) : data for test computation

        Returns: DataFrame with two columns, ``added`` and ``multiplied``.

        """

        preds = self.model.predict(X)
        preds = pd.DataFrame(dict(added=preds[:, 0],
                                  multiplied=preds[:, 1]))
        return preds

    def save_model(self, path=None):
        """Save the SampleModel.

        If no path is passed, tries to use model_path attribute.

        Args:
            path:

        Returns:

        """

        # save to user-specified, or model's path
        path = Path(path) if path else None
        save_path = path or self.model_path
        if save_path is None:
            raise FileNotFoundError("Must provide save_path")

        # create if necessary
        save_path.parent.mkdir(exist_ok=True)

        # keras' save
        self.model.save(save_path, include_optimizer=False)

    def load_data(self, data_path):
        """SampleModel loads pickled data

        Args:
            data_path:

        Returns:

        """

        with open(data_path, 'rb') as f:
            data = pickle.load(f)

        return data



from pathlib import Path

import pandas as pd


full = pd.read_csv("output-full.csv")
simp = pd.read_csv("output-full-simple.csv")

mapping = dict()

for x in p:
	x_name = x.name.replace("._", "")
	x_path = str(x).replace("._", "")

	if x.is_file():
		if x_name in mapping:
			mapping[x_name].append(x_path)
		else:
			mapping[x_name] = [x_path]

print(mapping)


from pathlib import Path
import matplotlib
import argparse
from collections import namedtuple
from multiprocessing.pool import ThreadPool
import os
import pickle
import time

import numpy as np
import pandas as pd
import skvideo.io
import scipy.misc

import tensorflow as tf
from tensorflow.python.keras.applications import ResNet50, InceptionV3, Xception, NASNetMobile, InceptionResNetV2
from tensorflow.python.keras.applications.resnet50 import preprocess_input as preprocess_input_resnet50
from tensorflow.python.keras.applications.xception import preprocess_input as preprocess_input_xception
from tensorflow.python.keras.applications.inception_v3 import preprocess_input as preprocess_input_inception_v3
from tensorflow.python.keras.applications.nasnet import preprocess_input as preprocess_input_nasnet
from tensorflow.python.keras.applications.inception_resnet_v2 import preprocess_input as preprocess_input_inception_resnet_v2
from tensorflow.python.keras.callbacks import ModelCheckpoint, TensorBoard
from tensorflow.python.keras.layers import Dense, Dropout
from tensorflow.python.keras.layers import GlobalMaxPooling2D, GlobalAveragePooling2D, AveragePooling2D
from tensorflow.python.keras.layers import Input
from tensorflow.python.keras.models import Model
from tensorflow.python.keras.optimizers import SGD, Adam, RMSprop
from tensorflow.python.keras.regularizers import l1
from tensorflow.python.keras import backend as K
from tqdm import tqdm

from .metrics import pri_matrix_loss
from zamba.models.cnnensemble.src import config
from zamba.models.cnnensemble.src import utils

# turn off noisy CPP compilation notifications
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
tf_config = tf.ConfigProto()
tf_config.gpu_options.allow_growth = True
# tf_config.gpu_options.visible_device_list = "0"
K.set_session(tf.Session(config=tf_config))

INPUT_ROWS = 404
INPUT_COLS = 720
INPUT_CHANNELS = 3
INPUT_SHAPE = (INPUT_ROWS, INPUT_COLS, INPUT_CHANNELS)
VIDEO_FPS = 24
PREDICT_FRAMES = [2, 8, 12, 18] + [i * VIDEO_FPS // 2 + 24 for i in range(14 * 2)]

from .config import CLASSES
NB_CLASSES = len(CLASSES)

cnnensemble_path = config.MODEL_DIR


def build_model_resnet50(lock_base_model: bool):
    base_model = ResNet50(input_shape=INPUT_SHAPE, include_top=False, pooling=None)
    if lock_base_model:
        for layer in base_model.layers:
            layer.trainable = False
    x = AveragePooling2D((5, 5), name='avg_pool5', strides=1)(base_model.layers[-2].output)
    x = GlobalMaxPooling2D()(x)
    res = Dense(NB_CLASSES, activation='sigmoid', name='classes', kernel_initializer='zero',
                kernel_regularizer=l1(1e-5))(x)
    model = Model(inputs=base_model.inputs, outputs=res)
    return model


def build_model_resnet50_avg(lock_base_model: bool):
    """
    Build the Resnet50 based level 1 model
    :param lock_base_model:
    :return:
    """
    base_model = ResNet50(input_shape=INPUT_SHAPE, include_top=False, pooling=None)
    if lock_base_model:
        for layer in base_model.layers:
            layer.trainable = False
    x = GlobalAveragePooling2D(name='avg_pool_final')(base_model.layers[-2].output)
    res = Dense(NB_CLASSES, activation='sigmoid', name='classes', kernel_initializer='zero',
                kernel_regularizer=l1(1e-5))(x)
    model = Model(inputs=base_model.inputs, outputs=res)
    return model


def build_model_xception_avg(lock_base_model: bool):
    base_model = Xception(input_shape=INPUT_SHAPE, include_top=False, pooling=None, weights=None)
    if lock_base_model:
        for layer in base_model.layers:
            layer.trainable = False
    x = GlobalAveragePooling2D(name='avg_pool_final')(base_model.layers[-1].output)
    res = Dense(NB_CLASSES, activation='sigmoid', name='classes', kernel_initializer='zero',
                kernel_regularizer=l1(1e-5))(x)
    model = Model(inputs=base_model.inputs, outputs=res)
    return model


def build_model_inception_v3_avg(lock_base_model: True):
    base_model = InceptionV3(input_shape=INPUT_SHAPE, include_top=False, pooling=None)
    if lock_base_model:
        for layer in base_model.layers:
            layer.trainable = False
    # base_model.summary()
    x = GlobalAveragePooling2D(name='avg_pool_final')(base_model.layers[-1].output)
    res = Dense(NB_CLASSES, activation='sigmoid', name='classes', kernel_initializer='zero',
                kernel_regularizer=l1(1e-5))(x)
    model = Model(inputs=base_model.inputs, outputs=res)
    # model.summary()
    return model


def build_model_inception_v2_resnet(lock_base_model: True):
    img_input = Input(shape=INPUT_SHAPE)
    base_model = InceptionResNetV2(input_tensor=img_input, include_top=False, pooling='avg')
    if lock_base_model:
        for layer in base_model.layers:
            layer.trainable = False
    # base_model.summary()
    res = Dense(NB_CLASSES, activation='sigmoid', name='classes', kernel_initializer='zero',
                kernel_regularizer=l1(1e-5))(base_model.layers[-1].output)
    model = Model(inputs=img_input, outputs=res)
    # model.summary()
    return model


def build_model_nasnet_mobile(lock_base_model: True):
    img_input = Input(shape=INPUT_SHAPE)
    base_model = NASNetMobile(input_tensor=img_input, include_top=False, pooling='avg')
    if lock_base_model:
        for layer in base_model.layers:
            layer.trainable = False
    res = Dense(NB_CLASSES, activation='sigmoid', name='classes', kernel_initializer='zero',
                kernel_regularizer=l1(1e-5))(base_model.layers[-1].output)
    model = Model(inputs=img_input, outputs=res)
    return model


ModelInfo = namedtuple('ModelInfo', ['factory', 'preprocess_input', 'unlock_layer_name', 'batch_size'])

MODELS = {
    'resnet50_initial': ModelInfo(
        factory=build_model_resnet50_avg,
        preprocess_input=preprocess_input_resnet50,
        unlock_layer_name='activation_22',
        batch_size=32
    ),
    'resnet50': ModelInfo(
        factory=build_model_resnet50,
        preprocess_input=preprocess_input_resnet50,
        unlock_layer_name='activation_22',
        batch_size=32
    ),
    'resnet50_avg': ModelInfo(
        factory=build_model_resnet50_avg,
        preprocess_input=preprocess_input_resnet50,
        unlock_layer_name='activation_22',
        batch_size=32
    ),
    'inception_v3': ModelInfo(
        factory=build_model_inception_v3_avg,
        preprocess_input=preprocess_input_inception_v3,
        unlock_layer_name='mixed4',
        batch_size=32
    ),
    'inception_v2_resnet': ModelInfo(
        factory=build_model_inception_v2_resnet,
        preprocess_input=preprocess_input_inception_resnet_v2,
        unlock_layer_name='activation_75',
        batch_size=16
    ),
    'xception_avg': ModelInfo(
        factory=build_model_xception_avg,
        preprocess_input=preprocess_input_xception,
        unlock_layer_name='block4_pool',
        batch_size=16
    ),
    'nasnet_mobile': ModelInfo(
        factory=build_model_nasnet_mobile,
        preprocess_input=preprocess_input_nasnet,
        unlock_layer_name='activation_15',
        batch_size=16
    ),
}

# extra model names used for different checkpoints, ideas/etc
MODELS['xception_avg_ch10'] = MODELS['xception_avg']
MODELS['inception_v2_resnet_ch10'] = MODELS['inception_v2_resnet']
MODELS['inception_v2_resnet_extra'] = MODELS['inception_v2_resnet']


class SingleFrameCNNDataset:
    """
    SingleFrameCNNDataset class prepares training and validation sample generators for training L1 models.
    """

    def __init__(self, fold, preprocess_input_func, batch_size,
                 validation_batch_size=1,
                 use_non_blank_frames=False):
        """
        :param fold: verification fold in 1..4 range, or -1 when model trained on the full dataset
        :param preprocess_input_func: pre-process function to convert input data to distribution
                used during imagenet based training
        :param batch_size: train batch size
        :param validation_batch_size: validation batch size
        :param use_non_blank_frames: if True, information about estimated probabilities for non blank frames
                is used so blank frame from non blank clip is less likely to be used.
        """
        self.validation_batch_size = validation_batch_size
        self.batch_size = batch_size
        self.combine_batches = 1  # combine multiple batches in generator for parallel processing
        self.preprocess_input_func = preprocess_input_func
        self.training_set_labels_ds_full = pd.read_csv(config.TRAINING_SET_LABELS)
        # self.loaded_files = set([fn for fn in os.listdir(config.RAW_VIDEO_DIR) if fn.endswith('.mp4')])
        self.loaded_files = set([fn + '.mp4' for fn in os.listdir(config.TRAIN_IMG_DIR)])

        self.training_set_labels_ds = self.training_set_labels_ds_full[
            self.training_set_labels_ds_full.filename.isin(self.loaded_files)]
        self.file_names = list(self.training_set_labels_ds.filename)
        self.training_set_labels_ds = self.training_set_labels_ds.set_index('filename')

        self.folds = pd.read_csv(cnnensemble_path / "input" / "folds.csv")
        train_clips = set(self.folds[self.folds.fold != fold].filename)
        test_clips = set(self.folds[self.folds.fold == fold].filename)

        self.train_clips = list(self.loaded_files.intersection(train_clips))
        self.test_clips = list(self.loaded_files.intersection(test_clips))

        self.test_clips = self.test_clips[:self.validation_steps() * validation_batch_size]
        self.pool = ThreadPool(8)

        print('train clips:', len(self.train_clips))
        print('test clips:', len(self.test_clips))

        self.train_clips_per_cat = {cls: [] for cls in range(NB_CLASSES)}  # cat id -> list of video_id
        self.y_map = {}
        for cls in range(NB_CLASSES):
            self.train_clips_per_cat[cls] = list(
                self.training_set_labels_ds[self.training_set_labels_ds[CLASSES[cls]] > 0.5].index.values)

        for cls in range(NB_CLASSES):
            print(CLASSES[cls], len(self.train_clips_per_cat[cls]))

        self.non_blank_frames = {}
        if use_non_blank_frames:
            for fn in ['resnet50_avg_1_non_blank.pkl',
                       'resnet50_2_non_blank.pkl',
                       'resnet50_avg_3_non_blank.pkl',
                       'resnet50_avg_4_non_blank.pkl']:
                data = pickle.load(open(cnnensemble_path / 'output/prediction_train_frames/' + fn, 'rb'))
                self.non_blank_frames.update(data)

    def train_steps_per_epoch(self):
        """
        Return the number of batches used for one epoch training.

        Calculated in a way to ensure comparable number of clips used per epoch regardless of batch size.
        """
        preprocess_batch_size = self.batch_size * self.combine_batches
        return int(len(self.train_clips) / 2 // preprocess_batch_size * preprocess_batch_size // self.batch_size)

    def validation_steps(self):
        """
        Return the number of batches used for validation.

        Calculated in a way to ensure comparable number of clips used regardless of batch size.
        """
        preprocess_batch_size = self.validation_batch_size * self.combine_batches
        return len(self.test_clips) // preprocess_batch_size * preprocess_batch_size // self.validation_batch_size

    def load_train_clip(self, video_id, offset=4, hflip=False):
        """
        Load the train tensor for the single train video clip.

        Id decoded jpeg file is found, it's used, otherwise video clip is decoded.

        :param video_id: train video clip id
        :param offset: frame number
        :param hflip: video if flipped horizontally if true
        :return: pre-processed X
        """
        images_dir = config.TRAIN_IMG_DIR
        video_dir = config.RAW_VIDEO_DIR
        classes = self.training_set_labels_ds.loc[[video_id]].as_matrix(columns=CLASSES)

        base_name = video_id[:-4]

        loaded = False
        try:
            fn = os.path.join(images_dir, base_name, f'{offset+1:04}.jpg')
            X = scipy.misc.imread(fn).astype(np.float32)
            loaded = True
        except FileNotFoundError:
            pass

        if not loaded:
            v = skvideo.io.vread(os.path.join(video_dir, video_id))
            X = np.zeros(INPUT_SHAPE)
            while offset >= 0:
                try:
                    X = v[(offset+1)*12]
                    break
                except IndexError:
                    offset -= 1
            if X.shape != INPUT_SHAPE:
                X = scipy.misc.imresize(X, size=(INPUT_ROWS, INPUT_COLS), interp='bilinear').astype(np.float32)
            X = X.astype(np.float32)
            del v

        if hflip:
            X = X[:, ::-1]
        # utils.print_stats('X', X)
        return self.preprocess_input_func(X), classes[0]

    def choose_train_video_id(self):
        """
            Policy of choosing video clip for training.
            Classes are partially normalised.

        :return: train video id
        """
        while True:
            r = np.random.random()
            if r < 0.6:
                return np.random.choice(self.train_clips)
            else:
                cls = np.random.randint(0, NB_CLASSES)
                count_threshold = np.random.choice([1000, 100, 10, 1], p=[0.69, 0.2, 0.1, 0.01])
                if len(self.train_clips_per_cat[cls]) >= count_threshold:
                    return np.random.choice(self.train_clips_per_cat[cls])

    def generate(self, verbose=False):
        """
        Train dataset generator

        :param verbose: print debug information if true
        """

        batch_size = self.batch_size
        X = np.zeros(shape=(batch_size,) + INPUT_SHAPE, dtype=np.float32)
        y = np.zeros(shape=(batch_size, NB_CLASSES), dtype=np.float32)

        def load_clip(video_id):
            if video_id in self.non_blank_frames:
                weights = self.non_blank_frames[video_id]
                offset = np.random.choice(list(range(1, 17)), p=weights / np.sum(weights))
            else:
                offset = np.random.randint(1, 17)
            return self.load_train_clip(video_id, offset=offset, hflip=np.random.choice([True, False]))

        while True:
            video_ids = self.train_clips
            np.random.shuffle(video_ids)
            # video_ids = video_ids[:self.train_steps_per_epoch() * batch_size]

            for i in range(int(self.train_steps_per_epoch() // self.combine_batches)):
                values_to_process = batch_size * self.combine_batches
                request_ids = [self.choose_train_video_id() for _ in
                               range(values_to_process)]  # video_ids[i*values_to_process: (i+1)*values_to_process]
                if verbose:
                    print(request_ids)
                results = self.pool.map(load_clip, request_ids)

                for j in range(values_to_process):
                    X[j % batch_size], y[j % batch_size] = results[j]
                    if (j + 1) % batch_size == 0:
                        yield X, y

    def generate_test(self, verbose=False, output_video_ids=False):
        batch_size = self.validation_batch_size
        X = np.zeros(shape=(batch_size,) + INPUT_SHAPE, dtype=np.float32)
        y = np.zeros(shape=(batch_size, NB_CLASSES), dtype=np.float32)

        def load_clip(video_id):
            if video_id in self.non_blank_frames:
                weights = self.non_blank_frames[video_id]
                offset = np.argmax(weights) + 1  # first frame is skipped
            else:
                offset = 4
            return self.load_train_clip(video_id, offset=offset)

        while True:
            video_ids = self.test_clips

            for i in range(int(self.validation_steps() // self.combine_batches)):
                values_to_process = batch_size * self.combine_batches
                request_ids = video_ids[i * values_to_process: (i + 1) * values_to_process]
                if verbose:
                    print(request_ids)
                results = self.pool.map(load_clip, request_ids)

                for j in range(values_to_process):
                    X[j % batch_size], y[j % batch_size] = results[j]
                    if (j + 1) % batch_size == 0:
                        yield X, y

    def frames_from_video_clip(self, video_fn):
        return self.preprocess_input_func(load_video_clip_frames(video_fn))

    def generate_frames_for_prediction(self):
        for video_id in sorted(self.test_clips):
            X = self.frames_from_video_clip(video_fn=os.path.join(config.RAW_VIDEO_DIR, video_id))
            y = self.training_set_labels_ds.loc[[video_id]].as_matrix(columns=CLASSES)
            yield video_id, X, y

    def generate_test_frames_for_prediction(self, data_path=None):
        if data_path is None:
            data_path = config.TEST_VIDEO_DIR
        test_ds = pd.read_csv(config.SUBMISSION_FORMAT)
        for video_id in test_ds.filename:
            X = self.frames_from_video_clip(video_fn=os.path.join(data_path, video_id))
            yield video_id, X

def load_video_clip_frames(video_fn):
    """
    Load video clip frames used for the second stage training and prediction.

    Returned frames matches PREDICT_FRAMES and resized if necessary to (INPUT_ROWS, INPUT_COLS)

    :param video_fn: path to video clip
    :return: ndarray of shape (PREDICT_FRAMES, INPUT_ROWS, INPUT_COLS, INPUT_CHANNELS)
    """
    X = np.zeros(shape=(len(PREDICT_FRAMES),) + INPUT_SHAPE, dtype=np.float32)

    v = skvideo.io.vread(str(video_fn))

    for i, frame_num in enumerate(PREDICT_FRAMES):
        try:
            frame = v[frame_num]
            if frame.shape != INPUT_SHAPE:
                frame = scipy.misc.imresize(frame, size=(INPUT_ROWS, INPUT_COLS), interp='bilinear').astype(
                    np.float32)
            else:
                frame = frame.astype(np.float32)
            X[i] = frame
        except IndexError:
            if i > 0:
                X[i] = X[i - 1]
            else:
                X[i] = 0.0
    return X


def check_generator(use_test):
    """
    Helper function used to visually check X and y vectors generated by SingleFrameCNNDataset
    :param use_test: true if validation samples are checked, false for training samples.
    """
    dataset = SingleFrameCNNDataset(preprocess_input_func=preprocess_input_resnet50,
                                    batch_size=2,
                                    validation_batch_size=2,
                                    fold=1)
    batch_id = 0
    startTime = time.time()

    if use_test:
        gen = dataset.generate_test()
    else:
        gen = dataset.generate()

    for X, y in gen:
        batch_id += 1
        elapsedTime = time.time() - startTime
        startTime = time.time()
        print(f'{batch_id} {elapsedTime:.3}')
        for batch_frame in range(dataset.batch_size):
            print(y[batch_frame])
            plt.imshow(utils.preprocessed_input_to_img_resnet(X[batch_frame]))
            plt.show()


def train(fold, model_name, weights='', initial_epoch=0, use_non_blank_frames=True):
    """
    Train the level 1 model.

    :param fold: Fold to train model on, or -1 to train on the full dataset
    :param model_name: name of the level 1 model to train
    :param weights: optional, if non empty string passed, training starts from supplied weights,
                    otherwise imagenet pre-trained model is loaded
    :param initial_epoch: Epoch to continue training from, used together with weight parameter to continue training
    :param use_non_blank_frames:
    :param use_extra_clips:
    :return: None
    """
    model_info = MODELS[model_name]
    dataset = SingleFrameCNNDataset(preprocess_input_func=model_info.preprocess_input,
                                    fold=fold,
                                    batch_size=model_info.batch_size,
                                    validation_batch_size=model_info.batch_size,
                                    use_non_blank_frames=use_non_blank_frames)

    model = model_info.factory(lock_base_model=True)
    if initial_epoch == 0 and weights == '':
        # train the first layer first unless continue training
        model.compile(optimizer=Adam(lr=1e-4), loss='binary_crossentropy', metrics=['accuracy'])
        model.fit_generator(
            dataset.generate(),
            steps_per_epoch=dataset.train_steps_per_epoch(),
            epochs=1,
            verbose=1)
    else:
        print('load weights', weights)
        model.load_weights(weights)

    utils.lock_layers_until(model, model_info.unlock_layer_name)
    # model.summary()

    checkpoints_dir = cnnensemble_path / f'output/checkpoints/{model_name}_fold_{fold}'
    tensorboard_dir = cnnensemble_path / f'output/tensorboard/{model_name}_fold_{fold}'
    os.makedirs(checkpoints_dir, exist_ok=True)
    os.makedirs(tensorboard_dir, exist_ok=True)

    checkpoint_periodical = ModelCheckpoint(checkpoints_dir + "/checkpoint-{epoch:03d}-{loss:.4f}-{val_loss:.4f}.hdf5",
                                            verbose=1,
                                            save_weights_only=True,
                                            period=1)
    tensorboard = TensorBoard(tensorboard_dir, histogram_freq=0, write_graph=False, write_images=False)

    # SGD with lr=1e-4 seems to be training very slowly, but still keep it for initial weights adjustments
    nb_sgd_epoch = 2
    if initial_epoch < nb_sgd_epoch:
        model.compile(optimizer=SGD(lr=1e-4, momentum=0.9), loss='binary_crossentropy', metrics=['accuracy'])
        model.fit_generator(
            dataset.generate(),
            steps_per_epoch=dataset.train_steps_per_epoch(),
            epochs=nb_sgd_epoch,
            verbose=1,
            validation_data=dataset.generate_test(),
            validation_steps=dataset.validation_steps(),
            callbacks=[
                checkpoint_periodical,
                tensorboard
            ],
            initial_epoch=initial_epoch
        )

    model.compile(optimizer=Adam(lr=5e-5), loss='binary_crossentropy', metrics=['accuracy'])
    model.fit_generator(
        dataset.generate(),
        steps_per_epoch=dataset.train_steps_per_epoch(),
        epochs=15,
        verbose=1,
        validation_data=dataset.generate_test(),
        validation_steps=dataset.validation_steps(),
        callbacks=[
            checkpoint_periodical,
            tensorboard
        ],
        initial_epoch=max(initial_epoch, nb_sgd_epoch)
    )

    model.save_weights(cnnensemble_path / f'output/{model_name}_s_fold_{fold}.h5')


def check_model(model_name, weights, fold):
    """
    Helper test method to do a simple sanity check of model predictions

    :param model_name: name of the model
    :param weights: model weights file path
    :param fold: fold to check
    """
    model = MODELS[model_name].factory(lock_base_model=True)
    model.load_weights(weights, by_name=False)

    dataset = SingleFrameCNNDataset(preprocess_input_func=MODELS[model_name].preprocess_input,
                                    batch_size=1,
                                    validation_batch_size=1,
                                    fold=fold)
    batch_id = 0
    for X, y in dataset.generate_test():
        pred = model.predict_on_batch(X)
        print()
        for i, cls in enumerate(CLASSES):
            print(f'gt: {y[0, i]}  pred: {pred[0, i]:.03f}  {cls}')
        batch_id += 1
        for batch_frame in range(dataset.batch_size):
            plt.imshow(utils.preprocessed_input_to_img_resnet(X[batch_frame]))
            # plt.imshow(X[batch_frame]/2.0+0.5)
            plt.show()


def generate_prediction(model_name, weights, fold):
    """
    Generate out of fold prediction for the training dataset, used for level 2 model training.

    Predictions are saved in individual files in the output/prediction_train_frames directory
    and need to be combined with save_combined_train_results()

    :param model_name: name of level 1 model
    :param weights: path to model weights
    :param fold: fold to generate predictions for
    :return:
    """
    model = MODELS[model_name].factory(lock_base_model=True)
    model.load_weights(weights, by_name=False)

    output_dir = cnnensemble_path / "output" / "prediction_train_frames" / f"{model_name}_{fold}"
    output_dir.mkdir(parents=True, exist_ok=True)

    dataset = SingleFrameCNNDataset(preprocess_input_func=MODELS[model_name].preprocess_input,
                                    batch_size=1,
                                    validation_batch_size=1,
                                    fold=fold)

    # skip processed files
    converted_files = set()
    processed_files = 0
    for video_id in dataset.test_clips:
        res_fn = output_dir.resolve() / f"{video_id}.csv"
        if res_fn.exists():
            processed_files += 1
            converted_files.add(video_id)
    test_clips = sorted(list(set(dataset.test_clips) - converted_files))

    def load_file(video_id):
        X = dataset.frames_from_video_clip(video_fn=os.path.join(config.RAW_VIDEO_DIR, video_id))
        y = dataset.training_set_labels_ds.loc[[video_id]].as_matrix(columns=CLASSES)
        return video_id, X, y

    start_time = time.time()

    pool = ThreadPool(8)
    prev_res = None
    for batch in utils.chunks(test_clips, 8, add_empty=True):
        if prev_res is not None:
            results = prev_res.get()
        else:
            results = []
        prev_res = pool.map_async(load_file, batch)
        for video_id, X, y in tqdm(results, desc='Processing videos...'):
            processed_files += 1
            res_fn = output_dir.resolve() / f"{video_id}.csv"
            have_data_time = time.time()
            prediction = model.predict(X, batch_size=1)

            ds = pd.DataFrame(index=[-1] + PREDICT_FRAMES,
                              data=np.row_stack([y, prediction]),
                              columns=CLASSES)
            ds.to_csv(res_fn, index_label='frame', float_format='%.5f')

            have_prediction_time = time.time()
            prepare_ms = int((have_data_time - start_time) * 1000)
            predict_ms = int((have_prediction_time - have_data_time) * 1000)
            start_time = time.time()
            # print(f'{video_id}  {processed_files} prepared in {prepare_ms} predicted in {predict_ms}')


def generate_prediction_test(model_name, weights, file_names, verbose=False, save_results=False):
    """
    Predict classes probabilities for a number of frames of each test clip

    :param model_name: name of the model
    :param weights: path to model weights
    :param verbose: print status messaged if true
    :param save_results: if set, prediction results are saved to
               output/prediction_test_frames/{model_name}_{fold}/{video_id}.csv

    :return: predictions as array with shape (nb_clips, nb_frames, nb_classes)
    """

    if verbose:
        print(model_name)

    K.clear_session()

    model = MODELS[model_name].factory(lock_base_model=True)
    model.load_weights(weights, by_name=False)

    output_dir = cnnensemble_path / "output" / "prediction_test_frames" / f"{model_name}"
    if save_results:
        output_dir.mkdir(parents=True, exist_ok=True)

    preprocess_input = MODELS[model_name].preprocess_input

    def load_file(video_id):
        try:
            X = preprocess_input(load_video_clip_frames(video_id))
        except:
            X = None

        return video_id, X

    start_time = time.time()

    all_predictions = []
    skipped_files = []
    processed_files = 0

    pool = ThreadPool(8)
    prev_res = None


    with tqdm(total=len(file_names), desc=f'Processing {len(file_names)} videos') as pbar:
        for batch in utils.chunks(file_names, 8, add_empty=True):
            if prev_res is not None:
                results = prev_res.get()
            else:
                results = []
            prev_res = pool.map_async(load_file, batch)

            for file_path, X in results:
                pbar.update(1)
                if X is None:
                    print(" Skipping file that is not a valid video: ", file_path)
                    processed_files += 1
                    skipped_files.append(file_path)

                else:
                    video_id = file_path.name
                    processed_files += 1
                    have_data_time = time.time()
                    prediction = model.predict(X, batch_size=1)
                    all_predictions.append(prediction)

                    if save_results:
                        res_fn = output_dir.resolve() / f"{video_id}.csv"
                        ds = pd.DataFrame(index=PREDICT_FRAMES,
                                          data=prediction,
                                          columns=CLASSES)
                        ds.to_csv(res_fn, index_label='frame', float_format='%.5f')

    return np.array(all_predictions), skipped_files


def find_non_blank_frames(model_name, fold):
    """
    For all non blank clips in the training dataset,
    estimate probability an animal is visible on particular video frames.

    This information is used to train next models, so it's less likely
    model is trained on frame labeled as non blank which is actually blank.

    Prediction is done for out of fold non blank frames only,
    so to cover the whole training dataset it's necessary to call find_non_blank_frames() for each fold.

    The result is saved to output/prediction_train_frames directory and loaded by DataSet class.

    :param model_name: name of the trained L1 model
    :param fold: verification fold model is trained with
    """
    data_dir = cnnensemble_path / f'output/prediction_train_frames/{model_name}_{fold}'
    training_labels = pd.read_csv(config.TRAINING_SET_LABELS)
    training_labels = training_labels.set_index('filename')

    res = {}

    for fn in tqdm(sorted(os.listdir(data_dir))):
        if not fn.endswith('csv'):
            continue
        filename = fn[:-len('.csv')]
        if training_labels.loc[filename].blank > 0.9:
            continue

        ds = np.loadtxt(os.path.join(data_dir, fn), delimiter=',', skiprows=1)
        target_frames = np.arange(0, 16) * 12 + 5
        src_frames = ds[1:, 0]

        blank_col = 2
        dst_blank_prob = np.interp(target_frames, src_frames, ds[1:, blank_col])
        res[filename] = 1.0 - dst_blank_prob
    pickle.dump(res, open(cnnensemble_path / f"output/prediction_train_frames/{model_name}_{fold}_non_blank.pkl", "wb"))


def save_combined_train_results(model_name, fold):
    X_raw = []
    y = []
    video_ids = []
    train_path = cnnensemble_path / f'output/prediction_train_frames/{model_name}_{fold}'
    raw_cache_fn = cnnensemble_path / f'output/prediction_train_frames/{model_name}_{fold}_combined.npz'
    for fn in tqdm(sorted(os.listdir(train_path))):
        if not fn.endswith('csv'):
            continue
        ds = np.loadtxt(os.path.join(train_path, fn), delimiter=',', skiprows=1)

        # top row is y, top col is frame number
        X_raw.append(ds[1:, 1:])
        y.append(ds[0, 1:])
        video_ids.append(fn[:-4])

    X_raw = np.array(X_raw)
    y = np.array(y)
    np.savez(raw_cache_fn, X_raw=X_raw, y=y, video_ids=video_ids)


def save_combined_test_results(model_name, fold, skip_existing=True):
    ds = pd.read_csv(config.SUBMISSION_FORMAT)
    data_dir = cnnensemble_path / 'output' / 'prediction_test_frames' / f'{model_name}_{fold}'
    res_fn = cnnensemble_path / 'output' / 'prediction_test_frames' / f'{model_name}_{fold}_combined.npy'

    if skip_existing and os.path.exists(res_fn):
        print('skip existing', res_fn)
        return

    X_raw = []
    for video_id in tqdm(ds.filename):
        ds = np.loadtxt(os.path.join(data_dir, video_id+'.csv'), delimiter=',', skiprows=1)
        # 0 col is frame number
        X_raw.append(ds[:, 1:])
    X_raw = np.array(X_raw).astype(np.float32)
    np.save(res_fn, X_raw)


def save_all_combined_train_results():
    for model in config.MODEL_WEIGHTS.keys():
        for fold in config.TRAIN_FOLDS:
            save_combined_train_results(model, fold)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='parallel cnn')
    parser.add_argument('action', type=str, default='check_model')
    parser.add_argument('--weights', type=str, default='')
    parser.add_argument('--initial_epoch', type=int, default=0)
    parser.add_argument('--fold', type=int, default=1)
    parser.add_argument('--model', type=str, default='resnet50')
    parser.add_argument('--use_non_blank_frames', action='store_true')
    parser.add_argument('--use_extra_clips', action='store_true')
    parser.add_argument('--file_names', type=str, nargs='+', required=False)

    args = parser.parse_args()
    action = args.action
    model = args.model

    if action == 'check_generator':
        check_generator(use_test=False)
    if action == 'check_generator_test':
        check_generator(use_test=True)
    elif action == 'check_model':
        check_model(model_name=model, weights=args.weights, fold=args.fold)
    elif action == 'train':
        train(fold=args.fold,
              model_name=model,
              weights=args.weights,
              initial_epoch=args.initial_epoch,
              use_non_blank_frames=args.use_non_blank_frames)
    elif action == 'generate_prediction':
        generate_prediction(fold=args.fold, model_name=model, weights=args.weights)
    elif action == 'generate_prediction_test':
        generate_prediction_test(model_name=model, weights=args.weights, file_names=args.file_names)
    elif action == 'find_non_blank_frames':
        find_non_blank_frames(fold=args.fold, model_name=model)
    elif action == 'save_combined_test_results':
        save_combined_test_results(fold=args.fold, model_name=model)
    elif action == 'save_all_combined_train_results':
        save_all_combined_train_results()



import os
from pathlib import Path
import pickle
from multiprocessing.pool import Pool

import numpy as np
from sklearn.model_selection import train_test_split

from xgboost import XGBClassifier

from tensorflow.python.keras.callbacks import ModelCheckpoint, TensorBoard, LearningRateScheduler
from tensorflow.python.keras.layers import Dense, Dropout
from tensorflow.python.keras.layers import Input
from tensorflow.python.keras.models import Model
from tensorflow.python.keras.optimizers import Adam
from tensorflow.python.keras.regularizers import l1

from zamba.models.cnnensemble.src import config, metrics, utils

import logging
logger = logging.getLogger(__file__)

NB_CAT = len(config.CLASSES)


def preprocess_x_histogram(data: np.ndarray):
    """
        Pre-process results of l1 model prediction to be used as an input to l2 model.

        This method is used for xgboost and lgbm models and uses set of per class statistics and histogram

        :param data: np.ndarray, predictions of the single l1 model with the shape of (NB_CLIPS, NB_FRAMES, NB_CLASSES)
        :return: pre-processed X tensor with shape (NB_CLIPS, X_WIDTH)
    """
    rows = []

    for row in data:
        items = [
            np.mean(row, axis=0),
            np.median(row, axis=0),
            np.min(row, axis=0),
            np.max(row, axis=0),
            np.percentile(row, q=10, axis=0),
            np.percentile(row, q=90, axis=0),
        ]
        for col in range(row.shape[1]):
            items.append(np.histogram(row[:, col], bins=10, range=(0.0, 1.0), density=True)[0])
        rows.append(np.hstack(items).flatten())

    return np.array(rows)


def preprocess_x_sorted_bins(data: np.ndarray):
    """
        Pre-process results of l1 model prediction to be used as an input to l2 model.

        This method is used for MLP model.

        Per class predictions are sorted and each bin of 4 values is averaged.

        :param data: np.ndarray, predictions of the single l1 model with the shape of (NB_CLIPS, NB_FRAMES, NB_CLASSES)
        :return: pre-processed X tensor with shape (NB_CLIPS, X_WIDTH)
    """
    rows = []
    for row in data:
        items = []
        for col in range(row.shape[1]):
            sorted = np.sort(row[:, col])
            items.append(sorted.reshape(-1, config.L2_SORTED_BINS_DOWNSAMPLE).mean(axis=1))
        rows.append(np.hstack(items))
    return np.array(rows)


def load_train_data(model_name, fold, preprocess_x):
    """
    Load the out of fold train data from file saved from l1 model prediction

    :param model_name: l1 model name
    :param fold: fold number NOT used for model training
    :param preprocess_x: function used to pre-process model input

    :return: X, y, video_ids
                np.ndarray with pre-processed train data;
                np.ndarray with labels as one hot encoded array
                list of video ids of X and y
    """
    data_path = config.MODEL_DIR / 'output/prediction_train_frames'
    raw_cache_fn = f'{data_path}/{model_name}_{fold}_combined.npz'
    cached = np.load(raw_cache_fn)
    X_raw, y, video_ids = cached['X_raw'], cached['y'], cached['video_ids']
    X = preprocess_x(X_raw)
    return X, y, video_ids


class SecondLevelModel:
    """
    Base class for the second level model, used to combine multiple individual l1 models predictions
    from multiple video frames.
    """
    def __init__(self, l1_model_names, preprocess_l1_model_output):
        """
        :param l1_model_names: list or l1 model names used this model combines results from
        :param preprocess_l1_model_output: function to pre-process l1 models output
        """
        self.l1_model_names = l1_model_names
        self.preprocess_l1_model_output = preprocess_l1_model_output
        self.pool = Pool(config.N_CORES)

    def _predict(self, X):
        """
        Internal, to be implemented by model implementation.

        Predict class probabilities

        :param X: Input vector
        :return: class probabilities as np.ndarray with shape (NB_ITEMS, NB_CLASSES)
        """
        raise NotImplemented()

    def _train(self, X, y):
        """
        Internal, to be implemented by model

        Train model.

        :param X:
        :param y: result as labels (not categorical) TODO: switch to categorical to allow multiple classes on one clip
        :return:
        """
        raise NotImplemented()

    def predict(self, l1_model_results):
        """
        :param l1_model_results: results of l1 model predictions as a dictionary
                             { model_name : ndarray(NB_CLIPS, NB_FRAMES, NB_CLASSES) }
        :return: class probabilities as ndarray with shape (NB_CLIPS, NB_CLASSES)
        """
        X_per_model = self.pool.map(self.preprocess_l1_model_output,
                                    [l1_model_results[model] for model in self.l1_model_names])
        X = np.column_stack(X_per_model)
        return self._predict(X)

    def train(self):
        """
        Train model using saved L1 model out of fold predictions
        """
        X_all_combined = []
        y_all_combined = []

        # list of (model_name, fold) tuples
        models_with_folds = []

        for model_name in self.l1_model_names:
            model_with_folds = []
            for fold in config.TRAIN_FOLDS:
                model_with_folds.append((model_name, fold))
            models_with_folds.append(model_with_folds)

        requests = []
        for model_with_folds in models_with_folds:
            for model_name, fold in model_with_folds:
                requests.append((model_name, fold, self.preprocess_l1_model_output))

        with utils.timeit_context('load all data'):
            results = self.pool.starmap(load_train_data, requests)

        for model_with_folds in models_with_folds:
            X_combined = []
            y_combined = []
            for model_name, fold in model_with_folds:
                X, y, video_ids = results[requests.index((model_name, fold))]
                logger.debug("Load", model_name, fold, X.shape, y.shape)
                X_combined.append(X)
                y_combined.append(y)

            X_all_combined.append(np.row_stack(X_combined))
            y_all_combined.append(np.row_stack(y_combined))

        X = np.column_stack(X_all_combined)
        y = y_all_combined[0]

        logger.debug("Training model", X.shape, y.shape)

        y_cat = np.argmax(y, axis=1)
        self._train(X, y_cat)


class SecondLevelModelXGBoost(SecondLevelModel):
    """
        XGBoost based L2 model implementation
    """
    def __init__(self, combined_model_name, l1_model_names):
        super().__init__(l1_model_names, preprocess_l1_model_output=preprocess_x_histogram)
        self.combined_model_name = combined_model_name
        self.model = None
        self.model_fn = config.MODEL_DIR / f"output/xgb_combined_{self.combined_model_name}.pkl"

    def _predict(self, X):
        if self.model is None:
            self.model = pickle.load(open(self.model_fn.resolve(), "rb"))
        return self.model.predict_proba(X)

    def _train(self, X, y):
        self.model = XGBClassifier(n_estimators=1600, objective='multi:softprob',
                                   learning_rate=0.03, silent=False, n_jobs=config.N_CORES)
        with utils.timeit_context('fit 1600 est'):
            self.model.fit(X, y)  # , eval_set=[(X_test, y_test)], early_stopping_rounds=20, verbose=True)
        pickle.dump(self.model, open(self.model_fn.resolve(), "wb"))


class SecondLevelModelMLP(SecondLevelModel):
    """
        Keras neural network based L2 model implementation
    """
    def __init__(self, combined_model_name, l1_model_names):
        super().__init__(l1_model_names, preprocess_l1_model_output=preprocess_x_sorted_bins)
        self.combined_model_name = combined_model_name
        self.model = None
        self.weights_fn = config.MODEL_DIR / f"output/nn_{self.combined_model_name}_full.pkl"

    def _build_model(self, input_size):
        input_data = Input(shape=(input_size,))
        x = input_data
        x = Dense(2048, activation='relu')(x)
        x = Dropout(0.75)(x)
        x = Dense(128, activation='relu')(x)
        x = Dense(NB_CAT, activation='sigmoid', kernel_regularizer=l1(1e-5))(x)
        model = Model(inputs=input_data, outputs=x)
        return model

    def _predict(self, X):
        if self.model is None:
            self.model = self._build_model(input_size=X.shape[1])
            self.model.load_weights(self.weights_fn.resolve())
            self.model.compile(optimizer=Adam(lr=1e-4), loss='binary_crossentropy', metrics=['accuracy'])
        return self.model.predict(X)

    def _train(self, X, y):
        self.model = self._build_model(input_size=X.shape[1])
        self.model.compile(optimizer=Adam(lr=1e-4), loss='binary_crossentropy', metrics=['accuracy'])
        self.model.summary()

        batch_size = 64

        def scheduler(epoch):
            if epoch < 32:
                return 1e-3
            if epoch < 48:
                return 4e-4
            if epoch < 80:
                return 1e-4
            return 1e-5

        self.model.fit(X, y,
                       batch_size=batch_size,
                       epochs=80,
                       verbose=1,
                       callbacks=[LearningRateScheduler(schedule=scheduler)])

        self.model.save_weights(self.weights_fn.resolve())


def predict(l1_model_results, profile):
    """
    :param l1_model_results: results of l1 model predictions as a dictionary
                             { model_name : ndarray(NB_CLIPS, NB_FRAMES, NB_CLASSES) }
    :return: class probabilities as ndarray with shape (NB_CLIPS, NB_CLASSES)
    """
    l1_model_names = config.PROFILES[profile]
    xgboost_model = SecondLevelModelXGBoost(l1_model_names=l1_model_names, combined_model_name='preset_'+profile)
    mlp_model = SecondLevelModelMLP(l1_model_names=l1_model_names, combined_model_name='preset_'+profile)

    xgboost_predictions = xgboost_model.predict(l1_model_results)
    mlp_predictions = mlp_model.predict(l1_model_results)

    return xgboost_predictions*0.5 + mlp_predictions*0.5



import numpy as np
import pandas as pd
import os
import shutil
import config
import matplotlib.pyplot as plt

predictions = [
    'resnet50_avg_2.csv',
    'inception_v3_1.csv',
    'inception_v3_2.csv',
    'xception_avg_1.csv',
    'xception_avg_2.csv'
]

data_dir = config.MODEL_DIR / 'output/prediction_unused_frames/'

classes = ['bird', 'blank', 'cattle', 'chimpanzee', 'elephant', 'forest buffalo',
           'gorilla', 'hippopotamus', 'human', 'hyena', 'large ungulate',
           'leopard', 'lion', 'other (non-primate)', 'other (primate)', 'pangolin',
           'porcupine', 'reptile', 'rodent', 'small antelope', 'small cat',
           'wild dog', 'duiker', 'hog']

classes_compatible = ['bird', 'blank', 'cattle', 'chimpanzee', 'elephant', 'forest buffalo',
                      'gorilla', 'hippopotamus', 'human', 'hyena', 'large ungulate',
                      'leopard', 'lion', 'other non-primate', 'other primate', 'pangolin',
                      'porcupine', 'reptile', 'rodent', 'small antelope', 'small cat',
                      'wild dog', 'duiker', 'hog']


def find_problematic_clips():
    datasets = [pd.read_csv(os.path.join(data_dir, fn)).sort_values(by=['filename']) for fn in predictions]

    ds0 = pd.merge(datasets[0], datasets[1], on='filename', suffixes=['0', '1'])
    ds1 = pd.merge(datasets[2], datasets[3], on='filename', suffixes=['2', '3'])
    ds2 = pd.merge(ds0, ds1, on='filename')
    ds = pd.merge(ds2, datasets[4], on='filename')

    print(datasets[0].shape, datasets[1].shape, ds.shape)

    combined = np.array(
        [ds.as_matrix(columns=['{}{}'.format(cls, suffix) for cls in classes]) for suffix in ['0', '1', '2', '3', '']])
    err = np.square(combined - np.mean(combined, axis=0, keepdims=True))
    err = np.sum(np.sum(err, axis=0), axis=1)
    max_values = np.max(combined, axis=0)
    mean_values = np.mean(combined, axis=0)

    ds['error'] = err
    for i, cls in enumerate(classes):
        ds['max_' + cls] = max_values[:, i]
        ds['mean_' + cls] = mean_values[:, i]

    print(np.sum(ds.error < 0.1))

    ds_matching = ds[ds.error < 0.1]
    ds_matching.to_csv(Path(__file__).parent.parent / 'output/unused_matching.csv',
                       index=False, float_format='%.7f',
                       columns=['filename'] + ['mean_' + cls for cls in classes],
                       header=['filename'] + classes)
    return
    ds_sorted = ds.sort_values(by=['error'], ascending=False)
    dest_dir = config.MODEL_DIR / 'output/to_label/'

    for cls in classes_compatible:
        os.makedirs(dest_dir+'res/'+cls, exist_ok=True)

    for i, (_, row) in enumerate(ds_sorted.iterrows()):
        if i > 2000:
            break
        mean_categories = row.as_matrix(['mean_' + cls for cls in classes])
        top_categories_idx = np.argsort(mean_categories)[::-1]

        prefixes = ''
        for cat_idx in top_categories_idx[:5]:
            prob = mean_categories[cat_idx]
            if prob < 0.05:
                break
            prefixes = prefixes + '{} {:.2f}_'.format(classes_compatible[cat_idx], prob)

        fn = '{:02} {}{}'.format(i, prefixes, row.filename)
        # print(i, row.error, row.filename, fn)
        cur_dest_dir = os.path.join(dest_dir, '{:02}'.format(i // 100))
        os.makedirs(cur_dest_dir, exist_ok=True)
        print(os.path.join(config.UNUSED_VIDEO_DIR, row.filename),
              os.path.join(cur_dest_dir, fn))
        shutil.copy(os.path.join(config.UNUSED_VIDEO_DIR, row.filename),
                    os.path.join(cur_dest_dir, fn))


def generate_labeled():
    data_dir = config.MODEL_DIR / 'input/extra_data'
    labels = {}  # video_id -> class_id
    for category_id, dir_name in enumerate(classes_compatible):
        for fn in os.listdir(os.path.join(data_dir, dir_name)):
            if fn.endswith('mp4'):
                labels[fn.split('_')[-1]] = category_id

    data = np.zeros((len(labels), len(classes)))
    file_names = []
    for fn, cls_id in labels.items():
        data[len(file_names), cls_id] = 1.0
        file_names.append(fn)

    res = pd.DataFrame(
        data={'filename': file_names}
    )
    for col, cls in enumerate(classes):
        res[cls] = data[:, col]

    res.to_csv(Path(__file__).parent.parent / 'output/unused_labeled.csv',
               index=False, float_format='%.7f', header=True)


find_problematic_clips()
generate_labeled()



from collections import deque
from os import remove
from pathlib import Path
from shutil import rmtree
import pandas as pd

from tensorflow.python.keras.utils import get_file
from tqdm import tqdm

from .model import Model
from .cnnensemble.src.single_frame_cnn import generate_prediction_test
from .cnnensemble.src import second_stage
from .cnnensemble.src import config


class CnnEnsemble(Model):
    def __init__(self, model_path, profile=config.DEFAULT_PROFILE, tempdir=None, download_region='us', verbose=False):
        # use the model object's defaults
        super().__init__(model_path, tempdir=tempdir)
        self.profile = profile

        self._download_weights_if_needed(download_region)
        self.verbose = verbose

    def load_data(self, data_path):
        """ Loads data and returns it in a format that can be used
            by this model. This is not recursive, and loads files that
            do not start with ".".

        Args:
            data_path: A path to the input data

        Returns:
            The data.
        """
        p = Path(data_path)
        file_names = [x for x in p.iterdir() if x.is_file() and not x.name.startswith(".")]
        return file_names

    def predict(self, file_names):
        """Predict class probabilities for each input, X

        Args:
            file_names: input data, list of video clip paths

        Returns:

        """

        l1_results = {}
        prof = config.PROFILES[self.profile]
        all_skipped = set()
        for l1_model in tqdm(prof, desc=f'Predicting on {len(prof)} L1 models'):
            l1_results[l1_model], skipped = generate_prediction_test(model_name=l1_model,
                                                                     weights=(Path(__file__).parent / 'cnnensemble' / 'output' /
                                                                     'checkpoints' / config.MODEL_WEIGHTS[l1_model]),
                                                                     file_names=file_names,
                                                                     verbose=self.verbose,
                                                                     save_results=False)

            all_skipped |= set([f.name for f in skipped])

        l2_results = second_stage.predict(l1_results, profile=self.profile)

        preds = pd.DataFrame(
            {
                'filename': [file_name.name for file_name in file_names if file_name.name not in all_skipped],
                **{cls: l2_results[:, i] for i, cls in enumerate(config.CLASSES)}
            },
            columns=['filename'] + config.CLASSES
        )

        return preds.set_index('filename')

    def fit(self, X, y):
        """Use the same architecture, but train the weights from scratch using
        the provided X and y.

        Args:
            X: training data
            y: training labels

        Returns:

        """
        pass

    def finetune(self, X, y):
        """Finetune the network for a different task by keeping the
        trained weights, replacing the top layer with one that outputs
        the new classes, and re-training for a few epochs to have the
        model output the new classes instead.

        Args:
            X:
            y:

        Returns:

        """
        pass

    def save_model(self):
        """Save the model weights, checkpoints, to model_path.
        """

    def _download_weights_if_needed(self, download_region):
        """Checks for directories containing the ensemble weights, downloads them if neccessary.

        The CnnEnsemble uses many GB of pretrained weights for prediction. When using the package for the first
        time, these weights (too heavy for pypi) need to be downloaded. We use the built in keras utility get_file to
        handle the download and extraction. This function leaves some hidden files behind after extraction,
        which we manually remove.

        After download, the zamba/zamba/models/cnnensemble/ directory will have three new directories:

        1. input - contains training fold splits from the original data, and stores formatting information for output
        2. output - contains models weights for all models used in the ensemble
        3. data_fast - contains cached training image preprocessing results
        """

        download_regions = ['us', 'eu', 'asia']
        if download_region not in download_regions:
            raise ValueError(f"download_region must be one of:\t{download_regions}")
        else:
            region_urls = {'us': 'https://s3.amazonaws.com/drivendata-public-assets/',
                           'eu': 'https://s3.eu-central-1.amazonaws.com/drivendata-public-assets-eu/',
                           'asia': 'https://s3-ap-southeast-1.amazonaws.com/drivendata-public-assets-asia/'}
            region_url = region_urls[download_region]

        # file names, paths
        fnames = ["input.tar.gz", "output.tar.gz", "data_fast.zip"]

        cache_dir = Path(__file__).parent
        cache_subdir = Path("cnnensemble")

        paths_needed = [cache_dir / cache_subdir / "input",
                        cache_dir / cache_subdir / "output",
                        cache_dir / cache_subdir / "data_fast"]

        # download and extract if needed
        for path_needed, fname in zip(paths_needed, fnames):
            if not path_needed.exists():
                origin = region_url + fname
                get_file(fname=fname, origin=origin, cache_dir=cache_dir, cache_subdir=cache_subdir,
                         extract=True)

                # remove the compressed file
                remove(cache_dir / cache_subdir / fname)

        # remove hidden files or dirs if present
        cnnpath = cache_dir / cache_subdir
        hidden_dirs = [pth for pth in cnnpath.glob("**/*") if pth.parts[-1].startswith("._") and pth.is_dir()]
        if hidden_dirs:
            deque(map(rmtree, hidden_dirs))
        hidden_files = [pth for pth in cnnpath.glob("**/*") if pth.parts[-1].startswith("._") and pth.is_file()]
        if hidden_files:
            deque(map(remove, hidden_files))



# avoid using a backend when not installed
# as a framework.
import matplotlib
matplotlib.use('Agg')

# import h5py first and ignore warnings which will go away in 2.7.2 (prevents annoying messages on start)
import warnings  # noqa: E402
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=FutureWarning)
    import h5py  # noqa: F401



import pandas as pd
import numpy as np
from sklearn.model_selection import StratifiedKFold
import config

import shutil
import os


def generate_folds():
    training_set_labels_ds_full = pd.read_csv(config.TRAINING_SET_LABELS)

    CLASSES = ['bird', 'blank', 'cattle', 'chimpanzee', 'elephant', 'forest buffalo', 'gorilla', 'hippopotamus', 'human',
               'hyena', 'large ungulate', 'leopard', 'lion', 'other (non-primate)', 'other (primate)', 'pangolin',
               'porcupine', 'reptile', 'rodent', 'small antelope', 'small cat', 'wild dog', 'duiker', 'hog']

    training_set_labels_ds_full['fold'] = 0

    data = training_set_labels_ds_full.as_matrix(columns=CLASSES + ['filename', 'fold'])

    cls = np.argmax(data[:, :-2], axis=1)

    skf = StratifiedKFold(n_splits=4, shuffle=True, random_state=42)

    split_idx = list(skf.split(data, cls))

    for fold, fold_split in enumerate(split_idx):
        items = fold_split[1]
        data[items, -1] = fold + 1
        print(fold)
        print(np.sum(data[items, :-2], axis=0))

    training_set_labels_ds_full['fold'] = data[:, -1]
    training_set_labels_ds_full.to_csv(Path(__file__).parent.parent / 'input/folds.csv', columns=['filename', 'fold'], index=False)


def move_non_train():
    training_set_labels_ds_full = pd.read_csv(config.TRAINING_SET_LABELS)
    train_files = set(training_set_labels_ds_full.filename)

    src_dir = config.MODEL_DIR / 'input/raw/'
    dst_dir = config.MODEL_DIR / 'input/raw_test/'
    skip_files = 0
    for fn in os.listdir(Path(__file__).parent.parent / 'input/raw/'):
        if fn not in train_files:
            skip_files += 1
            shutil.move(os.path.join(src_dir, fn), os.path.join(dst_dir, fn))
            if skip_files + 1 % 100 == 0:
                print(skip_files)
    print(skip_files)


def move_unused():
    test_set = pd.read_csv(config.SUBMISSION_FORMAT)
    test_files = set(test_set.filename)

    src_dir = config.MODEL_DIR / 'input/raw_test/'
    dst_dir = config.MODEL_DIR / 'input/raw_unused/'
    skip_files = 0
    for fn in os.listdir(Path(__file__).parent.parent / 'input/raw_test/'):
        if fn not in test_files:
            skip_files += 1
            shutil.move(os.path.join(src_dir, fn), os.path.join(dst_dir, fn))
            if skip_files + 1 % 100 == 0:
                print(skip_files)
    print(skip_files)

move_unused()



import numpy as np
from sklearn.metrics import log_loss


def pri_matrix_loss(y_true: np.ndarray, y_pred: np.ndarray):
    return log_loss(y_true.flatten(), y_pred.flatten(), eps=1e-15)  # * y_true.shape[1]



from pathlib import Path

MODEL_DIR = Path(__file__).parent.parent
RAW_VIDEO_DIR = MODEL_DIR / "input" / "raw"
TRAIN_IMG_DIR = MODEL_DIR / "data_fast" / "pri_matrix" / "train_img"
TEST_VIDEO_DIR = MODEL_DIR / "input" / "raw_test"
UNUSED_VIDEO_DIR = MODEL_DIR / "input" / "raw_unused"
UNUSED_IMG_DIR = MODEL_DIR / "input" / "raw_unused_img"
SMALL_VIDEO_DIR = MODEL_DIR / "input" / "small"
TRAINING_SET_LABELS = MODEL_DIR / "input" / "Pri-matrix_Factorization_-_Training_Set_Labels.csv"
SUBMISSION_FORMAT = MODEL_DIR / "input" / "Pri-matrix_Factorization_-_Submission_Format.csv"

CLASSES = ['bird', 'blank', 'cattle', 'chimpanzee', 'elephant', 'forest buffalo', 'gorilla', 'hippopotamus', 'human',
           'hyena', 'large ungulate', 'leopard', 'lion', 'other (non-primate)', 'other (primate)', 'pangolin',
           'porcupine', 'reptile', 'rodent', 'small antelope', 'small cat', 'wild dog', 'duiker', 'hog']

# number of cores used for multiprocessing pools, jobs etc
N_CORES = 8

# downsample bins of sorted predictions per class for frames for the second level model input
L2_SORTED_BINS_DOWNSAMPLE = 4

# L2 models are trained on OOF L1 model predictions, folds are listed here
TRAIN_FOLDS = [1, 2, 3, 4]

MODEL_WEIGHTS = {
    'inception_v2_resnet': 'inception_v2_resnet_ch10_fold_0/checkpoint-010-0.0324-0.0300.hdf5',
    'inception_v3': 'inception_v3_fold_0/checkpoint-009-0.0345-0.0331.hdf5',
    'nasnet_mobile': 'nasnet_mobile_fold_0/checkpoint-012-0.0329-0.0315.hdf5',
    'xception_avg': 'xception_avg_ch10_fold_0/checkpoint-009-0.0318-0.0399.hdf5'
}

PROFILES = {
    'fast': ['nasnet_mobile'],
    'full': ['nasnet_mobile', 'inception_v3', 'xception_avg', 'inception_v2_resnet']
}

DEFAULT_PROFILE = 'full'



import math
import pickle
import time
import random
from contextlib import contextmanager
import concurrent.futures
from queue import Queue
import numpy as np


def preprocessed_input_to_img_resnet(x):
    # Zero-center by mean pixel
    x = x.copy()
    x[:, :, 0] += 103.939
    x[:, :, 1] += 116.779
    x[:, :, 2] += 123.68

    # 'BGR' -> RGB
    img = x.copy()
    img[:, :, 0] = x[:, :, 2]
    img[:, :, 1] = x[:, :, 1]
    img[:, :, 2] = x[:, :, 0]
    return img / 255.0


@contextmanager
def timeit_context(name):
    startTime = time.time()
    yield
    elapsedTime = time.time() - startTime
    print('[{}] finished in {} ms'.format(name, int(elapsedTime * 1000)))


def chunks(l, n, add_empty=False):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l) // n * n + n - 1, n):
        if len(l[i:i + n]):
            yield l[i:i + n]
    if add_empty:
        yield []


def load_data(file_name):
    with open(file_name, 'rb') as f:
        data = pickle.load(f)
    return data


def save_data(data, file_name):
    pickle.dump(data, open(file_name, 'wb'))


def lock_layers_until(model, first_trainable_layer, verbose=False):
    found_first_layer = False
    for layer in model.layers:
        if layer.name == first_trainable_layer:
            found_first_layer = True

        if verbose and found_first_layer and not layer.trainable:
            print('Make layer trainable:', layer.name)
            layer.trainable = True

        layer.trainable = found_first_layer


def rand_or_05():
    if random.random() > 0.5:
        return random.random()
    return 0.5


def rand_scale_log_normal(mean_scale, one_sigma_at_scale):
    """
    Generate a distribution of value at log  scale around mean_scale

    :param mean_scale:  
    :param one_sigma_at_scale: 67% of values between  mean_scale/one_sigma_at_scale .. mean_scale*one_sigma_at_scale
    :return: 
    """

    log_sigma = math.log(one_sigma_at_scale)
    return mean_scale*math.exp(random.normalvariate(0.0, log_sigma))


def print_stats(title, array):
    print('{} shape:{} dtype:{} min:{} max:{} mean:{} median:{}'.format(
        title,
        array.shape,
        array.dtype,
        np.min(array),
        np.max(array),
        np.mean(array),
        np.median(array)
    ))


def limit_mem(K):
    K.get_session().close()
    cfg = K.tf.ConfigProto()
    cfg.gpu_options.allow_growth = True
    K.set_session(K.tf.Session(config=cfg))


def parallel_generator(orig_gen, executor):
    queue = Queue(maxsize=8)

    def bg_task():
        for i in orig_gen:
            # print('bg_task', i)
            queue.put(i)
        # print('bg_task', None)
        queue.put(None)

    task = executor.submit(bg_task)
    while True:
        value = queue.get()
        if value is not None:
            yield value
            queue.task_done()
        else:
            queue.task_done()
            break
    task.result()


def parallel_generator_nthread(orig_gen, executor):
    queue = Queue(maxsize=8)

    def bg_task():
        for i in orig_gen:
            # print('bg_task', i)
            queue.put(i)
        # print('bg_task', None)
        queue.put(None)

    task = executor.submit(bg_task)
    while True:
        value = queue.get()
        if value is not None:
            yield value
            queue.task_done()
        else:
            queue.task_done()
            break
    task.result()


def test_parallel_generator():
    def task(i):
        time.sleep(0.1)
        print('task', i)
        return i

    def orig_gen(n):
        for i in range(n):
            yield task(i)

    res_orig = []
    with timeit_context('orig gen'):
        for i in orig_gen(5):
            time.sleep(0.1)
            res_orig.append(i)

    executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)
    res_par = []
    with timeit_context('parallel gen'):
        for i in parallel_generator(orig_gen(5), executor):
            time.sleep(0.1)
            res_par.append(i)

    assert res_orig == res_par


if __name__ == '__main__':
    test_parallel_generator()



from pathlib import Path
from setuptools import setup, find_packages


def load_reqs(path):
    reqs = []
    with open(path, 'r') as f:
        for line in f.readlines():
            if line.startswith('-r'):
                reqs += load_reqs(line.split(' ')[1].strip())
            else:
                req = line.strip()
                if req and not req.startswith('#'):
                    reqs.append(req)
    return reqs


req_path = Path(__file__).parent / 'requirements.txt'
requirements = load_reqs(req_path)

long_description = open(Path(__file__).parent / 'README.md').read()

setup(
    name='zamba',
    version='0.1.5',
    description='Zamba is a tool to identify the species seen in camera trap videos from sites in central Africa.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='DrivenData',
    author_email='info@drivendata.org',
    url='http://zamba.drivendata.org',
    project_urls={
        'Homepage': 'http://zamba.drivendata.org',
        'Documentation': 'http://zamba.drivendata.org/docs/',
        'Source Code': 'https://github.com/drivendataorg/zamba',
        'DrivenData': 'http://drivendata.co'
    },
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',

        # Pick your license as you wish (should match "license" above)
         'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',

        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Multimedia :: Video',
    ],
    download_url='https://github.com/drivendataorg/zamba/archive/0.1.tar.gz',
    keywords=['deep learning', 'camera', 'africa', 'classifier'],
    python_requires='>=3.5',
    install_requires=requirements,
    extras_require={
        "cpu": ["tensorflow==1.7.0"],
        "gpu": ["tensorflow-gpu==1.7.0"]
    },
    entry_points={
        'console_scripts': [
            'zamba=zamba.cli:main',
        ],
    },
    packages=find_packages(exclude=['dist', 'docs', 'zamba.tests']),
)
